#!/usr/bin/python

########################################################################
# cert-query.py
# (c) by Michael Stroeder, michael.stroeder@inka.de
########################################################################

Version='0.5.3'

########################################################################
# CGI-BIN for querying the OpenSSL certificate DB
# - Aufruf ohne Parameter ergibt leeres Eingabeformular.
# Anmerkungen:
# - Die verwendeten NAME-Attribute der Eingabefelder entsprechen den
#   Bezeichnungen der Felder in openssl.cnf.
########################################################################

import sys, os, string, time, \
       pycacnf, cgiforms, htmlbase, charset, openssl

from pycacnf import opensslcnf, pyca_section

nsBaseUrl    = pyca_section.get('nsBaseUrl','/')
nsGetCertUrl = pyca_section.get('nsGetCertUrl','get-cert.py')
ScriptMethod = pyca_section.get('ScriptMethod','POST')
HelpUrl      = pyca_section.get('HelpUrl',nsBaseUrl)

##############################################################################
# haeufig gebrauchte Funktionen
##############################################################################

# String der Hilfe-URL zu bestimmtem Parameter zurueckgeben
def HelpURL(name,text):
  return '<A HREF="%s/ns-enroll-help.html#%s">%s</A>' % (HelpUrl,name,charset.trans(text,charset.iso,charset.html))

# Ausdrucken eines leeren Eingabeformulars
def PrintEmptyForm(form,method='POST'):
  print '<FORM ACTION="%s" METHOD="%s">\n<TABLE>' % \
        (os.environ.get('SCRIPT_NAME','cert-query.py'),method)
  for i in form.keys:
    print '<TD WIDTH=20%%>%s:</TD><TD>%s</TD></TR>' % (HelpURL(form.field[i][0].name,form.field[i][0].text),form.field[i][0].inputfield())
  print '</TABLE>'
  print '<INPUT TYPE="submit" VALUE="Search">'
  print '<INPUT TYPE="reset" VALUE="Reset">'
  print '</FORM>'
  return

# Ausgabe der Ergebnistabelle
def PrintFound(form,found,cellpadding=5,width=100):
  # key der Checkbox entfernen, da kein Feld
  searchkeys=form.keys
  try:
    searchkeys.remove('casesensitive')
  except:
    pass
  try:
    searchkeys.remove('onlyvalid')
  except:
    pass

  print '<TABLE BORDER CELLPADDING=%d%% WIDTH=%d%%>' % (cellpadding,width)
  # Tabellenueberschriften ausgeben
  print '<TR><TH>Serial</TH><TH>valid<BR>until</TH>'
  for i in searchkeys:
    print '<TH><FONT SIZE=-1>%s</FONT></TH>' % (HelpURL(i,form.field[i][0].text))
  print '</TR>'

  # Tabelleninhalte 
  for ca_name in found.keys():
    for i in found[ca_name]:
      # Eine Tabellenzeile ausgeben
      print '<TR>'
      print '<TD><A HREF="%s%s/%s/email?%s">%s</A></TD>' % (nsBaseUrl,nsGetCertUrl,ca_name,i[openssl.db.DB_serial],i[openssl.db.DB_serial])
      print '<TD>%s</TD>' % time.strftime('%d.%m.%Y',time.localtime(time.mktime(openssl.db.dbtime2tuple(i[openssl.db.DB_exp_date]))))
      dnfield = openssl.db.SplitDN(i[openssl.db.DB_name])
      # Spaltenelemente ausgeben
      for j in form.keys:
	if dnfield.has_key(j) and dnfield[j]:
          print '<TD><FONT SIZE=-1>%s</FONT></TD>' % charset.trans(dnfield[j],charset.t61,charset.html)
	else:
          # bei leeren Feldern Leerzeichen, damit Tabelle immer Raender hat
          print '<TD>&nbsp;</TD>'
      print '</TR>'

  print '</TABLE>'

  return

##############################################################################
# Hauptprogramm 
##############################################################################

# form initialisieren
form = cgiforms.formClass()

# Die gueltigen Inputattribute setzen
alphanumregex = '[0-9a-zA-Z.*?_ -]*'
mailadrregex = '[0-9a-zA-Z@.*?=/_ -]*'

form.add(cgiforms.formInputClass('C','Country',2,'[a-zA-Z?]'*2))
form.add(cgiforms.formInputClass('ST','State / Province',30,alphanumregex))
form.add(cgiforms.formInputClass('L','Location',30,alphanumregex))
form.add(cgiforms.formInputClass('O','Organization',30,alphanumregex))
form.add(cgiforms.formInputClass('OU','Organizational Unit',30,alphanumregex))
form.add(cgiforms.formInputClass('CN','Common Name',30,alphanumregex))
form.add(cgiforms.formInputClass('Email','E-Mail',40,mailadrregex))
form.add(cgiforms.formCheckboxClass('casesensitive','case sensitive','yes'))
form.add(cgiforms.formCheckboxClass('onlyvalid','only valid','yes',1))

# Schon Parameter vorhanden?
if not form.contentlength:

  # Aufruf erfolgte ohne Parameter =>
  # 0. Schritt: leeres Eingabeformular ausgeben

  htmlbase.PrintHeader('Search certificates')
  htmlbase.PrintHeading('Search certificates')
  print """You can search for valid certificates in the
certificate database.<P>Just type in substrings or
regular expressions as search criteria."""
  PrintEmptyForm(form)
  htmlbase.PrintFooter()
  sys.exit(0)

# Aufruf erfolgte mit Parametern
try:
  form.getparams()
except cgiforms.formContentLengthException,e:
  htmlbase.PrintErrorMsg('Content length invalid.')
  sys.exit(0)
except cgiforms.formParamNameException,e:
  htmlbase.PrintErrorMsg('Unknown parameter "%s".' % (e.name))
  sys.exit(0)
except cgiforms.formParamContentException,e:
  htmlbase.PrintErrorMsg('Content of field "%s" has invalid format.' % (e.text))
  sys.exit(0)
except cgiforms.formParamStructException,e:
  htmlbase.PrintErrorMsg('Too many (%d) parameters for field "%s".' % (e.count,e.name))
  sys.exit(0)
except cgiforms.formParamLengthException,e:
  htmlbase.PrintErrorMsg('Content too long. Field "%s" has %d characters.' % (e.text,e.length))
  sys.exit(0)
except:
  htmlbase.PrintErrorMsg('Unknown exception.')
  sys.exit(0)

# Parameter sind ok

# keys der Checkboxen entfernen, falls vorhanden, da kein Suchparameter
searchkeys=form.inputkeys
try:
  searchkeys.remove('casesensitive')
except:
  pass
try:
  searchkeys.remove('onlyvalid')
except:
  pass

# Anfrage zurecht formatieren
query = openssl.db.empty_DN_dict

for i in searchkeys:
  query[i] = form.field[i][0].content

found ={}
old_db_filenames = []

for ca_name in opensslcnf.sectionkeys['ca']:

  ca_section = opensslcnf.data[opensslcnf.data['ca'][ca_name]]
  
  # Ist der Zertifikattyp 'S/MIME for client use' ?
  if opensslcnf.isemailcert(ca_name):

    # Stammverzeichnis der CA
    ca_dir = ca_section.get('dir','')
    ca_database = string.replace(ca_section.get('database','$dir/index.txt'),'$dir',ca_dir)
    # Zertifikat-DB schon vorher mal behandelt?
    if not ca_database in old_db_filenames:
      old_db_filenames.append(ca_database)
      if os.path.isfile(ca_database):
	# Anfrage starten
	found[ca_name] = openssl.db.GetEntriesbyDN(ca_database,query,form.field['casesensitive'][0].content=='yes',form.field['onlyvalid'][0].content=='yes')

# Nix gefunden!
if not found:
  htmlbase.PrintErrorMsg('No matching entries found.')
  sys.exit(0)

# Ausgabe des Suchergebnisses
htmlbase.PrintHeader('Search results')
PrintFound(form,found)
htmlbase.PrintFooter()

sys.exit(0)
