#!/usr/bin/python

########################################################################
# ns-enroll.py 
# (c) by Michael Stroeder, michael.stroeder@inka.de
########################################################################

Version='0.5.3'

########################################################################
# CGI-BIN for creating certificate requests with Netscape Navigator
# or compatible browsers which are aware of <KEYGEN>
# - Calling the CGI-BIN without parameters gives back an
#   empty input form
########################################################################

import os, sys, types, string, \
       whrandom, smtplib, \
       pycacnf, \
       cgiforms, charset, ipadr, openssl, htmlbase

from pycacnf import opensslcnf, pyca_section

########################################################################
# "Internet"-Funktionen
########################################################################

# Aus einer Mailadresse die Domain herausloesen
# Ergebnis ist Teilstring hinter letztem(!) @
# falls kein @ vorhanden, dann Leerstring
def DomainAdr(MailAdr=""):
  splitted=string.split(MailAdr,'@')
  if len(splitted)>1:
    return splitted[-1]
  else:
    return ''

# Mailadresse mittels DNS-Abfrage auf Gueltigkeit ueberpruefen.
# Ergebnis ist 1, falls ok
def CheckMailAdr(MailAdr=""):
#  domain = DomainAdr(MailAdr)
#  # Resolver anschmeissen
#  resolver=DNS.ParseResolvConf()
#  r = DNS.DnsRequest(name=domain)
#  a = r.req(qtype='A')
#  if not len(a.answers):
#    mx = r.req(qtype='MX')
#    return (mx!=[])
  return 1

# String der Hilfe-URL zu bestimmtem Parameter zurueckgeben
def HelpURL(HelpUrlBase,name,text):
  return '<A HREF="%sns-enroll-help.html#%s">%s</A>' % \
         (HelpUrlBase,name,charset.trans(text,charset.iso,charset.html))

# Ausgabe eines leeren Eingabeformulars
def PrintEmptyForm(form,ca_name,policy,req,method='POST'):
  print '<FORM ACTION="%s/%s" METHOD=%s>' % (os.environ.get('SCRIPT_NAME','ns-enroll.py'),ca_name,ScriptMethod)
  print '<TABLE WIDTH=100% BORDER>'
  for i in form.keys:
    for j in form.field[i]:
      print '<TR><TD>%s%s</TD><TD>%s</TD></TR>' % (HelpURL(HelpUrlBase,j.name,j.text),' <FONT SIZE="+2">*</FONT>' * j.required,j.inputfield())
  print '</TABLE>'
  print '<INPUT TYPE="submit" VALUE="Send">'
  print '<INPUT TYPE="reset" VALUE="Reset">'
  print '</FORM>'
  return

def PrintInput(form,cellpadding=5,width=100):
  print '<TABLE BORDER CELLPADDING=%d%% WIDTH=%d%%>' % (cellpadding,width)
  formKeys = form.inputkeys[:]
  for key in form.inputkeys:
    if form.field[key][0].content=='':
      formKeys.remove(key)  

  try:
    formKeys.remove('userpassword')
  except ValueError:
    pass
  try:
    formKeys.remove('challenge')
  except ValueError:
    pass

  for i in formKeys:
    for j in form.field[i]:
      print '<TR><TD WIDTH=35%%>%s</TD><TD>%s</TD></TR>' % \
             (HelpURL(HelpUrlBase,j.name,j.text),j.contentprint())
  print '</TABLE>'
  return

# Ausgabe eines Formulars zur Schuesselerzeugung
def PrintKeygenForm(form,ca_name,ca):
  PrintInput(form) 
  print '<FORM ACTION="%s/%s" METHOD=%s>' % (os.environ.get('SCRIPT_NAME','ns-enroll.py'),ca_name,ScriptMethod)
  for i in form.inputkeys:
    for j in form.field[i]:
      print '<INPUT TYPE="hidden" NAME="%s" VALUE="%s">' % (j.name,j.content)
  if ca.min_key_size>0:
    print """Please note:<BR>
The certificate type <STRONG>%s</STRONG> requires a minimum key size of <STRONG>%d</STRONG> bits!
If you are not able to choose a key length equal or greater than <STRONG>%d</STRONG> the
certificate authority will refuse to issue a certificate for your certificate request!<P>
""" % (ca_name,ca.min_key_size,ca.min_key_size)
  print '%s:%s<P><INPUT TYPE="submit" VALUE="Generate key pair"></FORM>' % (HelpURL(HelpUrlBase,form.field['SPKAC'][0].name,form.field['SPKAC'][0].text),form.field['SPKAC'][0].inputfield(form.field['challenge'][0].content))

# **********************************************************
# ********************* Hauptprogramm **********************
# **********************************************************

MailRelay           = pyca_section.get('MailRelay','localhost')
TmpDir              = pyca_section.get('TmpDir','/tmp')
caCertReqMailAdr    = pyca_section.get('caCertReqMailAdr',DomainAdr(os.environ.get('SERVER_ADMIN','cert-req@localhost')))
caPendCertReqValid  = string.atoi(pyca_section.get('caPendCertReqValid','0'))

caInternalCertTypes = pyca_section.get('caInternalCertTypes',[])
if type(caInternalCertTypes)!=types.ListType:
  caInternalCertTypes = [caInternalCertTypes]

caInternalIPAdr     = pyca_section.get('caInternalIPAdr',['127.0.0.1/255.255.255.255'])
if type(caInternalIPAdr)!=types.ListType:
  caInternalIPAdr = [caInternalIPAdr]

caInternalDomains   = pyca_section.get('caInternalDomains','')
if type(caInternalDomains)!=types.ListType:
  caInternalDomains = [caInternalDomains]

ScriptMethod        = pyca_section.get('ScriptMethod','POST')

# Ein paar Umgebungsvariablen auslesen, welche der Apache liefert
remote_addr = os.environ.get('REMOTE_ADDR','')
http_user_agent = os.environ.get('HTTP_USER_AGENT','')
ca_name = os.environ.get('PATH_INFO','')[1:]

ca_names = opensslcnf.sectionkeys.get('ca',[])

if not ca_names:
  htmlbase.PrintErrorMsg('No certificate authorities found.')
  sys.exit(0)

if not ca_name:
  htmlbase.PrintErrorMsg('No certificate authority.<BR>You might want to have a look at the <A HREF="ca-index.py">list of certificate types.</A>')
  sys.exit(0)

if not ca_name in ca_names:
  # CA-Definition nicht in openssl-Konfiguration enthalten
  htmlbase.PrintErrorMsg('Unknown certificate authority "%s".' % ca_name)
  sys.exit(0)

ca = opensslcnf.getcadata(ca_name)

HelpUrlBase = ca.nsBaseUrl+pyca_section.get('HelpUrl','')

policy_section = opensslcnf.data.get(ca.policy,{})
req_section = opensslcnf.data.get(ca.req,{})

if req_section and req_section.has_key('distinguished_name'):
  req_distinguished_name_section = opensslcnf.data.get(req_section['distinguished_name'],{})
  req_distinguished_name_keys = opensslcnf.sectionkeys.get(req_section['distinguished_name'],[])
else:
  htmlbase.PrintErrorMsg('Request section for "%s" not found.' % ca_name)
  sys.exit(0)

# Hier Verwendungszweck der Zertifikate prfen
if not opensslcnf.isclientcert(ca_name):
  htmlbase.PrintErrorMsg('Certificate authority "%s" does not issue client certificates.' % ca_name)
  sys.exit(0)

# form initialisieren
form = cgiforms.formClass()

# Die gueltigen Inputattribute setzen
alphanumregex = r'[0-9a-zA-Z/\'"._ -]*'
# telephoneregex = r'^\+[0-9][0-9]-[0-9]*-[0-9]*'
form.add(cgiforms.formPasswordClass('challenge','Initial Master Secret',30,alphanumregex,required=1))

# Vorgaben fuer certificate requests aus openssl.cnf beachten
dn_attr_keys = []
dn_attr = {}
for i in req_distinguished_name_keys:
  l = string.split(i,'_')
  attr_name = string.strip(l[0])
  if not attr_name in dn_attr_keys:
    dn_attr_keys.append(attr_name)
    dn_attr[attr_name]={'comment':'','max':'40','regex':alphanumregex,'default':''}
  if len(l)>1:
    dn_attr[attr_name][l[1]]=req_distinguished_name_section.get(i,'')
  elif len(l)==1:
    dn_attr[attr_name]['comment']=req_distinguished_name_section.get(i,attr_name)

for i in dn_attr_keys:
  imaxlength=string.atoi(dn_attr[i].get('max','40'))
  if imaxlength>40:
    isize=40
  else:
    isize=imaxlength
  policy_field = policy_section.get(i,'optional')
  if policy_field=='match':
    form.add(cgiforms.formHiddenInputClass(i,dn_attr[i]['comment'],imaxlength,dn_attr[i]['regex'],dn_attr[i]['default'],required=1,show=1))
  else:
    if type(dn_attr[i]['default'])==types.ListType:
      dn_attr[i]['default'].sort()
      form.add(cgiforms.formSelectClass(i,dn_attr[i]['comment'],dn_attr[i]['default'],required=policy_field=='supplied'))
    else:
      form.add(cgiforms.formInputClass(i,dn_attr[i]['comment'],imaxlength,dn_attr[i]['regex'],dn_attr[i]['default'],required=policy_field=='supplied',size=isize))

#form.add(cgiforms.formInputClass('streetaddress','Street',40,alphanumregex))
#form.add(cgiforms.formInputClass('postalcode','ZIP-Code',10,alphanumregex))
#form.add(cgiforms.formInputClass('telephoneNumber','Phone',40,telephoneregex))
#form.add(cgiforms.formInputClass('facsimileTelephoneNumber','FAX',40,telephoneregex))
#form.add(cgiforms.formInputClass('contactPerson','Contact Person',40,alphanumregex))
#form.add(cgiforms.formPasswordClass('userpassword','Password',30,alphanumregex))
#form.add(cgiforms.formPasswordClass('userpassword','Password (repeat)',30,alphanumregex))

# Schon Parameter vorhanden?
if not form.contentlength:

  # Aufruf erfolgte ohne Parameter =>
  # 0. Schritt: leeres Eingabeformular ausgeben

  # Noch mal den Browser checken
  if http_user_agent[:len('Mozilla')]!='Mozilla':
    htmlbase.PrintErrorMsg('This key enrollment only works with Netscape browsers.')
    sys.exit(0)

  if not ca.nsComment:
    ca.nsComment = 'No comment'
  if ca.nsCaPolicyUrl:
    nsCommentStr = '<A HREF="%s%s">%s</A>' % (ca.nsBaseUrl,ca.nsCaPolicyUrl,ca.nsComment)
  else:
    nsCommentStr = ca.nsComment

  htmlbase.PrintHeader('Input form for certificate request')
  htmlbase.PrintHeading('Input form for certificate request')
  print """<TABLE>
<TR><TD>Certificate authority:</TD><TD><STRONG>%s</STRONG></TD></TR>
<TR><TD>Certificate type:</TD><TD><STRONG>%s</STRONG></TD></TR>
<TR><TD>Certificate comment:</TD><TD><STRONG>%s</STRONG></TD></TR>
</TABLE><P>
""" % (ca_name,ca.nsCertTypeStr,nsCommentStr)
  print """You can apply for a certificate by filling out the input form below.
Click on the names of the parameters to get further informations about the
usage and format restrictions of the input data.<P>
Required input parameters are marked with *.
"""
  PrintEmptyForm(form,ca_name,policy_section,req_distinguished_name_section)
  htmlbase.PrintFooter()
  sys.exit(0)

# 1. und 2. Schritt haben Schluesselfeld
form.add(cgiforms.formKeygenClass('SPKAC','Key length',500))

# Aufruf erfolgte mit Parametern
try:
  form.getparams()
except cgiforms.formContentLengthException,e:
  htmlbase.PrintErrorMsg('Content length invalid.')
  sys.exit(0)
except cgiforms.formParamNameException,e:
  htmlbase.PrintErrorMsg('Unknown parameter "%s".' % (e.name))
  sys.exit(0)
except cgiforms.formParamNameMissing,e:
  htmlbase.PrintErrorMsg('Parameter "%s" is missing.' % (e.text))
  sys.exit(0)
except cgiforms.formParamContentException,e:
  htmlbase.PrintErrorMsg('Content of field "%s" has invalid format.' % (e.text))
  sys.exit(0)
except cgiforms.formParamStructException,e:
  htmlbase.PrintErrorMsg('Too many (%d) parameters for field "%s".' % (e.count,e.name))
  sys.exit(0)
except cgiforms.formParamLengthException,e:
  htmlbase.PrintErrorMsg('Content too long. Field "%s" has %d characters.' % (e.text,e.length))
  sys.exit(0)
#except:
#  htmlbase.PrintErrorMsg('Unknown exception.')
#  sys.exit(0)

##############################################################################
# Zusaetzliche Ueberpruefungen diverser Parameter
##############################################################################

commonName   = form.field.get('commonName',[])[0].content
emailAddress = form.field.get('emailAddress',[])[0].content

# E-Mail-Domain im DNS?
if not CheckMailAdr(emailAddress):
  htmlbase.PrintErrorMsg('Domain of e-mail address not found in DNS.<BR>Please provide your correct and valid %s or ask your systems administrator.'% \
                          HelpURL(HelpUrlBase,'emailAddress','E-Mail address'))
  sys.exit(0)

if "userpassword" in form.inputkeys:
  # Passwortvergleich
  if len(form.field["userpassword"])!=2 or \
     form.field["userpassword"][0].content!=form.field["userpassword"][1].content:
    # Es wurde nur ein Passwort angegeben oder
    # die beiden Passwoerter waren nicht gleich
    htmlbase.PrintErrorMsg('Passwords do not match.')
    sys.exit(0)

if (ca_name in caInternalCertTypes):
  if not (DomainAdr(emailAddress) in caInternalDomains):
    htmlbase.PrintErrorMsg('This type of certificate request is restricted to internal address domains!')
    sys.exit(0)

  if not ipadr.MatchIPAdrList(remote_addr,caInternalIPAdr):
    htmlbase.PrintErrorMsg('This type of certificate request is restricted to internal hosts!')
    sys.exit(0)

if not ('SPKAC' in form.inputkeys):

  # Aufruf erfolgte mit Parametern ohne Schluessel =>
  # 1. Schritt: Eingegebene Daten anzeigen und
  # Benutzer zur Schluesselerzeugung auffordern

  htmlbase.PrintHeader('Create key pair and certificate request')
  htmlbase.PrintHeading('Create key pair and certificate request')
  print 'Your key pair and certificate request can be generated now.<BR>'
  print 'Please have a look below to check if your input data was correct.<P>'
  PrintKeygenForm(form,ca_name,ca)
  htmlbase.PrintFooter()
  sys.exit(0)


# Aufruf erfolgte mit Parametern inkl. Schluessel =>
# 2. Schritt: Forminhalt bearbeiten

# Check the required key length if min_key_size was defined
if ca.min_key_size>0:
  # FIX ME!!!
  # This is a very primitive and falsy key length checking!!!
  minbytes={512:200,768:300,1024:400}
  if minbytes.has_key(ca.min_key_size) and len(form.field["SPKAC"][0].content)<minbytes[ca.min_key_size]:
    htmlbase.PrintErrorMsg('The key length you submitted was too weak!<BR>The certificate type <STRONG>%s</STRONG> requires a minimum key size of <STRONG>%d</STRONG> bits!' % (ca_name,ca.min_key_size))
    sys.exit(0)

# Gebraucht wird tatsaechlich nur ein Passwort
if "userpassword" in form.inputkeys:
  # Aus zwei mach eins
  form.field["userpassword"] = [form.field["userpassword"][0]]
    
# Zufaellige ChallengeID erzeugen und daraus eindeutige, noch nicht
# existierende Dateinamen basteln, pruefen!!!
zaehler=0
alreadyexists=1
while (zaehler<20) and alreadyexists:
  # Zufaellige ID fuer Antwort vom Benutzer
  caChallengeId    = whrandom.randint(11111111,99999999)
  # ca_name und ChallengeId fuer Mail-Subjects und Dateinamen
  camailSubject    = 'ns-cert-req.%s.%d' % (ca_name,caChallengeId)
  request_filename = os.path.join(ca.pend_reqs_dir,'%s.spkac' % (camailSubject))
  ldif_filename    = os.path.join(ca.pend_reqs_dir,'%s.ldif' % (camailSubject))
  zaehler = zaehler+1
  alreadyexists = os.path.exists(request_filename) or \
                  os.path.exists(ldif_filename)

if alreadyexists:
  # Alle Versuche nicht existierende Dateinamen zu basteln schlugen fehl
  htmlbase.PrintErrorMsg('Error generating a random ID or creating temporary files.')
  sys.exit(0)

##############################################################################
# LDAP-Server-Eintrag
##############################################################################

# Mapping von SPKAC- zu LDAP-Attributnamen
spkac2ldifname = {
  'countryName':'c',
  'stateOrProvinceName':'st',
  'localityName':'l',
  'organizationName':'o',
  'organizationalUnitName':'ou',
  'commonName':'cn',
  'emailAddress':'mail',
}

# DN fuer LDAP-Eintrag zusammenbauen:
# Besteht aus commonName und emailAddress, das sollte eindeutig genug sein.
# Ist organizationName angegeben, so kommt das zum DN dazu

if ('organizationName' in form.inputkeys) and form.field['organizationName'][0].content:
  ldap_dn   = charset.iso2utf('cn=%s,mail=%s,o=%s' % \
        	(form.field['commonName'][0].content,
		 form.field['emailAddress'][0].content,
		 form.field['organizationName'][0].content
		)
	      )
else:
  ldap_dn   = charset.iso2utf('cn=%s,mail=%s' % \
        	(form.field['commonName'][0].content,
		 form.field['emailAddress'][0].content
		)
	      )

# Schlssel kommt nicht in der Form ins LDAP
formKeys = form.inputkeys[:]
formKeys.remove('SPKAC')
# LDAP-Daten fr ldapcl.fileClass zusammenbauen
ldap_data = {}

# Liste der Dictionary der Form {attr:[entry]} aufbauen
ldap_data['objectclass'] = ['top','person','inetorgperson','strongauthenticationuser']
for i in formKeys:
  ldap_attr = spkac2ldifname.get(i,i)
  ldap_data[ldap_attr] = []
  for j in form.field[i]:
    if j.content:
      ldap_data[ldap_attr].append(charset.iso2utf(j.content))

# LDAP-Datei schreiben
ldap_file = open(ldif_filename,'w')
ldap_file.write('dn: mail=%s\n' % ldap_data['mail'][0])
for attr in ldap_data.keys():
  for value in ldap_data[attr]:
    ldap_file.write('%s: %s\n' % (attr,value))
ldap_file.write('\n')
ldap_file.close()
os.chmod(ldif_filename,0444)

# ldap_file = ldapcl.fileClass(ldif_filename)
# ldap_file.WriteAddRecord(ldap_dn,ldap_data)
  
##############################################################################
# SPKAC-Request erzeugen
##############################################################################

request_file = open(request_filename,'w') 
# Nur die wirklich im Zertifikat enthaltenen Parameter
CertRequestKeys = ['countryName','stateOrProvinceName','localityName','organizationName','organizationalUnitName','commonName','initials','uid','emailAddress','SPKAC']
for i in CertRequestKeys:
  if (i in form.inputkeys) and form.field[i][0].content:
    request_file.write('%s = %s\n' % (i,form.field[i][0].content))
request_file.close()
os.chmod(request_filename,0444)

##############################################################################
# Nette Mail fuer Challenge an Benutzer zusammenstellen
##############################################################################

if caCertReqMailAdr and emailAddress:

  if commonName:
    to_addr = '%s <%s>' % (commonName,emailAddress)
  else:
    to_addr = '%s' % (emailAddress)
  
  # Mailbody erzeugen
  mail_msg = """From: %s
To: %s
Subject: %s

Someone (maybe you) has sent a certificate request
to our certificate authority.

Please answer this e-mail with the same subject
if this was really you and the data below is correct.

If someone abused your name / e-mail address just
ignore this message. %s

------------- ns-cert-req -------------
""" % (caCertReqMailAdr,\
       to_addr,\
       camailSubject,\
       (caPendCertReqValid>0)*(' The certificate request will be\nremoved automatically after %d hours.' % caPendCertReqValid)
      )
  # Hier den eigentlichen Cert-Req an Mailbody anhaengen

  formKeys = form.keys[:len(form.keys)]
  try:
    formKeys.remove('userpassword')
  except ValueError:
    pass
  try:
    formKeys.remove('challenge')
  except ValueError:
    pass
  try:
    formKeys.remove('SPKAC')
  except:
    pass

  mail_msg_paramlist = []
  for i in formKeys:
    if (i in form.inputkeys) and form.field[i][0].content:
      mail_msg_paramlist.append('%s = %s' % (i,form.field[i][0].content))
  mail_msg = '%s%s' % (mail_msg,string.join(mail_msg_paramlist,'\n'))
  smtpconn=smtplib.SMTP(MailRelay)
  smtpconn.set_debuglevel(0)
  try:
    smtpconn.sendmail(caCertReqMailAdr,[to_addr],mail_msg)
  except:
    htmlbase.PrintErrorMsg('Unable to send an e-mail to <B>%s</B>!<BR>Please provide your correct and valid %s or ask your systems administrator.' % \
                            (to_addr,HelpURL(HelpUrlBase,'emailAddress','e-mail address')))
    # FIX ME!!! Eigentlich hier noch die temporren Dateien lschen!!!
    sys.exit(0)
  smtpconn.quit()

# Schliesslich und endlich...
htmlbase.PrintHeader('Certificate request stored.')
htmlbase.PrintHeading('Certificate request is stored.')
print """Your certificate request was stored for
further processing."""

if caCertReqMailAdr and emailAddress:
  print """<P>You will get an e-mail message with a random ID.
Please answer this e-mail to confirm your certificate request."""

if caPendCertReqValid:
  print """Otherwise your certificate request will be
removed automatically after %d hour(s).<P>""" % (caPendCertReqValid)

print '<P>Once again the data you gave to us:<P>'
PrintInput(form)
htmlbase.PrintFooter()

sys.exit(0)
