#######################################################################
# openssl.cert.py Version 0.5.3
# (c) by Michael Stroeder, michael.stroeder@propack-data.de
########################################################################
# Module for accessing certificate files
########################################################################

import string, os

import openssl

from openssl.cert import *

def GetCertValues(certfilename,inform=''):

  if not inform:
    ext = string.lower(os.path.splitext(certfilename)[1])
    if ext in ['pem','der','txt']:
      inform = ext
    else:
      inform = 'pem'

  pipe = os.popen('%s x509 -in %s -fingerprint -inform %s -noout -subject -issuer -dates -serial' % (openssl.bin_filename,certfilename,inform))

  result={}

  s=pipe.readline()
  while s:
    name, value = string.split(s,'=',1)
    result[name]=string.strip(value)
    s=pipe.readline()
  pipe.close()
  
  return result

class X509CertificateClass:

  def __init__(self,certfilename,inform=''):
    certattrs = GetCertValues(certfilename,inform)
    self.subject = openssl.db.SplitDN(certattrs.get('subject',openssl.db.empty_DN_dict))
    self.issuer = openssl.db.SplitDN(certattrs.get('issuer',openssl.db.empty_DN_dict))
    self.serial  = string.atoi(certattrs.get('serial','-1'))
    self.hash  = certattrs.get('hash','')
    self.notBefore = certattrs.get('notBefore','')
    self.notAfter = certattrs.get('notAfter','')
    # Fingerprint suchen
    certattrs_keys = certattrs.keys()
    self.fingerprint = ''
    self.digest = ''
    for i in certattrs_keys:
      if i[-11:]=='Fingerprint':
        self.fingerprint = certattrs[i]
        self.digest = string.split(i)[0]

  def serial(self):
    return string.atoi(self.serial)

