#!/usr/bin/python

########################################################################
# ca-certreq-mail.py
# (c) by Michael Stroeder, michael.stroeder@inka.de
########################################################################

Version = '0.5.3'

########################################################################
# Dieses Skript nimmt eine Mail auf Stanardeingabe mit einem cert-req
# an und speichert diese als Datei in einem Verzeichnis
# zur weiteren Verarbeitung
########################################################################

import sys, os, time, string, regex, rfc822, \
       pycacnf, openssl
       
from pycacnf import opensslcnf, pyca_section

ca_names = opensslcnf.sectionkeys.get('ca',[])
MailRelay           = pyca_section.get('MailRelay','localhost')
TmpDir              = pyca_section.get('TmpDir','/tmp')
caCertReqMailAdr    = pyca_section.get('caCertReqMailAdr','')
caInternalCertTypes = pyca_section.get('caInternalCertTypes',[])
caInternalIPAdr     = pyca_section.get('caInternalIPAdr',['127.0.0.1/255.255.255.255'])
caInternalDomains   = pyca_section.get('caInternalDomains','')
if type(caInternalDomains)!=type([]):
  caInternalDomains = [caInternalDomains]

# Name der Protokolldatei, falls '' dann wird stderr verwendet
logfilename='/tmp/ca-cert-mail-challenge'

###########################################

# Einen Datensatz in Protokolldatei schreiben
# log		Filehandle von bereits geoeffneter Protokolldatei
# Kategorie	des Eintrags, z.B. 'Error:'
# Mail		Mailheader
# Kommentar	klar
def LogWrite(log,Kategorie,Mail,Kommentar):
  log.write('%s %s: ' % (time.strftime('%d.%m.%Y %X',time.localtime(time.time())),Kategorie))
  for i in ['from','subject','message-id']:
    if Mail.has_key(i):
      log.write('%s ' % (Mail[i]))
  log.write('%s\n' % (Kommentar))
  return

# Antwortmail schicken
def Reply(MailAdr,Body):
  return

#############################################################
# Hauptprogramm
#############################################################

m=rfc822.Message(sys.stdin)

# Protokolldatei oeffnen
if logfilename:
  logfile = open(logfilename,'a')
else:
  logfile = sys.stderr

# Erst einmal relevante Parameter auf Gueltigkeit pruefen

# Laenge des Subjects gueltig?
if len(m["Subject"])>100:
  LogWrite(logfile,'Error',m,'Subject too long.')
  sys.exit(0)

#############################################################
# Format von Subject und Body ueberpruefen
# Wirkt recht paranoid, soll aber gegen Muellmails schuetzen
#############################################################

# Ueberlange Subjects verbieten
if len(m["subject"])>80:
  LogWrite(logfile,'Error',m,'Subject too long.')
  sys.exit(0)

subject = string.strip(m["subject"])
subjectstart = string.find(subject,'ns-cert-req.')

# Format dreiteilig?
try:
  prefix,ca_name,caChallengeId = string.split(subject[subjectstart:len(subject)],'.',2)
except ValueError:
  LogWrite(logfile,'Error',m,'Subject has wrong format.')
  sys.exit(0)

# Prfix richtig?
if prefix!='ns-cert-req':
  LogWrite(logfile,'Error',m,'Subject has wrong format.')
  sys.exit(0)

# ChallengeID sechsstellig und nur Ziffern?
if len(caChallengeId)!=8 or \
  regex.match('[0-9]'*8,caChallengeId)!=8:
  LogWrite(logfile,'Error',m,'caChallengeId %s has bad format.' % (caChallengeId))
  sys.exit(0)

# CA Name gltig?
if not (ca_name in ca_names):
  LogWrite(logfile,'Error',m,'ca_name "%s" wrong.' % (ca_name))
  sys.exit(0)

ca_section=opensslcnf.data[opensslcnf.data['ca'][ca_name]]
ca_dir = ca_section.get('dir','')

ca_pend_reqs_dir = string.replace(ca_section.get('pend_reqs_dir','$dir/pendreqs'),'$dir',ca_dir)
ca_new_reqs_dir  = string.replace(ca_section.get('new_reqs_dir','$dir/newsreqs'),'$dir',ca_dir)

# Eine Benutzerantwort ist eingetroffen
request_filename = os.path.join(ca_pend_reqs_dir,'%s.%s.%s' % (prefix,ca_name,caChallengeId))

# Existieren die benoetigten Dateien?
if not os.path.isfile('%s.spkac' % (request_filename)):
  LogWrite(logfile,'Error',m,'SPKAC file not found for caChallengeId %s.' % (caChallengeId))
  sys.exit(0)
if not os.path.isfile('%s.ldif' % request_filename):
  LogWrite(logfile,'Error',m,'LDIF file not found for caChallengeId %s.' % (caChallengeId))
  sys.exit(0)

# Hier sind jetzt alle Angaben gueltig, soweit pruefbar

# FIX ME!!! The ownership should be changed!
# os.chown('%s.spkac' % (request_filename),90,90)
# os.chown('%s.ldif' % (request_filename),90,90)

# Move files
newrequest_filename = os.path.join(ca_new_reqs_dir,'%s.%s.%s' % (prefix,ca_name,caChallengeId))
os.rename('%s.spkac' % (request_filename),'%s.spkac' % (newrequest_filename)) 
os.rename('%s.ldif' % (request_filename),'%s.ldif' % (newrequest_filename)) 

LogWrite(logfile,'Challenge',m,'Request challenge: %s Id=%s' % (ca_name,caChallengeId))

sys.exit(0)
